#######################################################################
#                            Flux methods                             #
#######################################################################



@with_signature function flux(equation::EulerEquation)
  @accepts State(rho, q, E)
  @unpack eos = equation
  eps, = internal_energy((rho,q,E), equation)
  p = eos(Pressure, Density, rho, InternalEnergy, eps)
  flx_rho = q
  flx_q   = q^2 / rho + p
  flx_E   = q / rho * (E + p)
  @returns flx_rho, flx_q, flx_E
end



@with_signature function speed(equation::EulerEquation)
  @accepts State(rho, q, E)
  @unpack eos = equation
  eps, = internal_energy((rho,q,E), equation)
  p = eos(Pressure, Density, rho, InternalEnergy, eps)
  v = abs(q / p) + sqrt(abs(eos.gamma * p / rho))
  @returns v
end


@with_signature function internal_energy(::EulerEquation)
  @accepts rho, q, E
  eps = E / rho - q^2 / (2 * rho^2)
  @returns eps
end


@with_signature function pressure(equation::EulerEquation{EquationOfState.IdealGas})
  @accepts rho, eps
  @unpack eos = equation
  p = eos(Pressure, Density, rho, InternalEnergy, eps)
  @returns p
end


@with_signature function entropy(equation::EulerEquation{EquationOfState.IdealGas})
  @accepts rho, p, x
  @unpack eos  = equation
  gamma = eos.gamma
  S = rho / (gamma - 1) * log(p / rho^gamma)
  @returns S
end


@with_signature function entropy_flux(equation::EulerEquation)
  @accepts rho, q, S
  @unpack eos  = equation
  flx_S = q / rho * S
  @returns flx_S
end


@with_signature function llf(equation::EulerEquation)
  @accepts Prefix(lhs), rho, flx_rho, q, flx_q, E, flx_E, v
  @accepts Prefix(rhs), rho, flx_rho, q, flx_q, E, flx_E, v
  @accepts nx

  vmax = absmax(lhs_v, rhs_v)

  lhs_nflx = nx*lhs_flx_rho
  rhs_nflx = nx*rhs_flx_rho
  nflx_rho = LLF(lhs_nflx, rhs_nflx, lhs_rho, rhs_rho, vmax)

  lhs_nflx = nx*lhs_flx_q
  rhs_nflx = nx*rhs_flx_q
  nflx_q = LLF(lhs_nflx, rhs_nflx, lhs_q, rhs_q, vmax)

  lhs_nflx = nx*lhs_flx_E
  rhs_nflx = nx*rhs_flx_E
  nflx_E = LLF(lhs_nflx, rhs_nflx, lhs_E, rhs_E, vmax)

  @returns nflx_rho, nflx_q, nflx_E
end


@with_signature function bdryllf(equation::EulerEquation)
  @accepts Prefix(lhs), rho, flx_rho, q, flx_q, E, flx_E, v
  @accepts Prefix(rhs), init_rho, init_flx_rho, init_q, init_flx_q, init_E, init_flx_E, init_v
  @accepts nx

  # enforce dirichlet/neumann conditions following stability analysis
  # see dg-notes/tex/evm.pdf
  # Hesthaven, Numerical Methods for Conservation Laws uses same logic

  # Dirichlet conditions
  rhs_rho = -lhs_rho + 2*rhs_init_rho
  rhs_q   = -lhs_q   + 2*rhs_init_q
  rhs_E   = -lhs_E   + 2*rhs_init_E

  # Neumann conditions
  # rhs_rho = lhs_rho
  # rhs_q   = lhs_q
  # rhs_E   = lhs_E

  @unpack eos = equation
  rhs_eps, = internal_energy((rhs_rho,rhs_q,rhs_E), equation)
  rhs_p = eos(Pressure, Density, rhs_rho, InternalEnergy, rhs_eps)
  rhs_flx_rho = rhs_q
  rhs_flx_q   = rhs_q^2 / rhs_rho + rhs_p
  rhs_flx_E   = rhs_q / rhs_rho * (rhs_E + rhs_p)

  vmax = absmax(lhs_v, rhs_init_v)

  lhs_nflx = nx*lhs_flx_rho
  rhs_nflx = nx*rhs_flx_rho
  nflx_rho = LLF(lhs_nflx, rhs_nflx, lhs_rho, rhs_rho, vmax)

  lhs_nflx = nx*lhs_flx_q
  rhs_nflx = nx*rhs_flx_q
  nflx_q = LLF(lhs_nflx, rhs_nflx, lhs_q, rhs_q, vmax)

  lhs_nflx = nx*lhs_flx_E
  rhs_nflx = nx*rhs_flx_E
  nflx_E = LLF(lhs_nflx, rhs_nflx, lhs_E, rhs_E, vmax)

  @returns nflx_rho, nflx_q, nflx_E
end


@with_signature function av_flux(eq::EulerEquation)
  @accepts flx_rho, ldg_rho, flx_q, ldg_q, flx_E, ldg_E, smoothed_mu

  flx_rho += smoothed_mu * ldg_rho
  flx_q   += smoothed_mu * ldg_q
  flx_E   += smoothed_mu * ldg_E

  @returns flx_rho, flx_q, flx_E, flx_E
end
@with_signature function av_nflux(eq::EulerEquation)
  @accepts Prefix(lhs), ldg_rho, ldg_q, ldg_E, smoothed_mu
  @accepts Prefix(rhs), ldg_rho, ldg_q, ldg_E, smoothed_mu
  @accepts nflx_rho, nflx_q, nflx_E, nx

  lhs_nflx = nx*lhs_ldg_rho
  rhs_nflx = nx*rhs_ldg_rho
  nflx_rho += 0.5 * (lhs_smoothed_mu*lhs_nflx + rhs_smoothed_mu*rhs_nflx)

  lhs_nflx = nx*lhs_ldg_q
  rhs_nflx = nx*rhs_ldg_q
  nflx_q  += 0.5 * (lhs_smoothed_mu*lhs_nflx + rhs_smoothed_mu*rhs_nflx)

  lhs_nflx = nx*lhs_ldg_E
  rhs_nflx = nx*rhs_ldg_E
  nflx_E += 0.5 * (lhs_smoothed_mu*lhs_nflx + rhs_smoothed_mu*rhs_nflx)

  @returns nflx_rho, nflx_q, nflx_E
end


@with_signature function ldg_nflux(eq::EulerEquation)
  @accepts Prefix(lhs), rho, q, E
  @accepts Prefix(rhs), rho, q, E
  @accepts nx

  nflx_rho = 0.5 * nx * (lhs_rho + rhs_rho)
  nflx_q   = 0.5 * nx * (lhs_q + rhs_q)
  nflx_E   = 0.5 * nx * (lhs_E + rhs_E)

  @returns nflx_rho, nflx_q, nflx_E
end
@with_signature function ldg_bdryllf(eq::EulerEquation)
  @accepts Prefix(lhs), rho, q, E
  @accepts Prefix(rhs), rho, q, E
  @accepts nx

  # set exterior state to 0 according to stability analysis
  # see dg-notes/tex/evm.pdf for reasoning
  nflx_rho = nx * lhs_rho
  nflx_q   = nx * lhs_q
  nflx_E   = nx * lhs_E

  @returns nflx_rho, nflx_q, nflx_E
end


#######################################################################
#                                 2D                                  #
#######################################################################


@with_signature function internal_energy(::EulerEquation2d)
  @accepts rho, qx, qy, E
  q2 = qx^2 + qy^2
  eps = E / rho - q2 / (2 * rho^2)
  @returns eps
end


@with_signature function pressure(equation::EulerEquation2d)
  @accepts rho, qx, qy, E
  q = sqrt(qx^2+qy^2)
  eps, = internal_energy((rho,qx,qy,E), equation)
  @unpack eos = equation
  p = eos(Pressure, Density, rho, InternalEnergy, eps)
  @returns p
end


@with_signature function flux(equation::EulerEquation2d)
  @accepts rho, qx, qy, E, p
  @unpack eos = equation
  q = sqrt(qx^2+qy^2)
  eps, = internal_energy((rho,qx,qy,E), equation)
  p = eos(Pressure, Density, rho, InternalEnergy, eps)
  flx_rho_x = qx
  flx_rho_y = qy
  flx_qx_x = qx^2/rho + p
  flx_qx_y = qx*qy / rho
  flx_qy_y = qy^2/rho + p
  flx_E_x  = qx / rho * (E + p)
  flx_E_y  = qy / rho * (E + p)
  @returns flx_rho_x, flx_rho_y, flx_qx_x, flx_qx_y, flx_qy_y, flx_E_x, flx_E_y
end


# TODO Replace with llf(::EulerEquation2d) below
@with_signature function llf_rho(equation::EulerEquation2d)
  @accepts Prefix(lhs), rho, flx_rho_x, flx_rho_y, v
  @accepts Prefix(rhs), rho, flx_rho_x, flx_rho_y, v
  @accepts nx, ny
  lhs_nflx = nx*lhs_flx_rho_x + ny*lhs_flx_rho_y
  rhs_nflx = nx*rhs_flx_rho_x + ny*rhs_flx_rho_y
  vmax = absmax(lhs_v, rhs_v)
  nflx_rho = LLF(lhs_nflx, rhs_nflx, lhs_rho, rhs_rho, vmax)
  @returns nflx_rho
end
@with_signature function llf_qx(equation::EulerEquation2d)
  @accepts Prefix(lhs), qx, flx_qx_x, flx_qx_y, v
  @accepts Prefix(rhs), qx, flx_qx_x, flx_qx_y, v
  @accepts nx, ny
  lhs_nflx = nx*lhs_flx_qx_x + ny*lhs_flx_qx_y
  rhs_nflx = nx*rhs_flx_qx_x + ny*rhs_flx_qx_y
  vmax = absmax(lhs_v, rhs_v)
  nflx_qx = LLF(lhs_nflx, rhs_nflx, lhs_qx, rhs_qx, vmax)
  @returns nflx_qx
end
@with_signature function llf_qy(equation::EulerEquation2d)
  @accepts Prefix(lhs), qy, flx_qy_x, flx_qy_y, v
  @accepts Prefix(rhs), qy, flx_qy_x, flx_qy_y, v
  @accepts nx, ny
  lhs_nflx = nx*lhs_flx_qy_x + ny*lhs_flx_qy_y
  rhs_nflx = nx*rhs_flx_qy_x + ny*rhs_flx_qy_y
  vmax = absmax(lhs_v, rhs_v)
  nflx_qy = LLF(lhs_nflx, rhs_nflx, lhs_qy, rhs_qy, vmax)
  @returns nflx_qy
end
@with_signature function llf_E(equation::EulerEquation2d)
  @accepts Prefix(lhs), E, flx_E_x, flx_E_y, v
  @accepts Prefix(rhs), E, flx_E_x, flx_E_y, v
  @accepts nx, ny
  lhs_nflx = nx*lhs_flx_E_x + ny*lhs_flx_E_y
  rhs_nflx = nx*rhs_flx_E_x + ny*rhs_flx_E_y
  vmax = absmax(lhs_v, rhs_v)
  nflx_E = LLF(lhs_nflx, rhs_nflx, lhs_E, rhs_E, vmax)
  @returns nflx_E
end


@with_signature function llf(equation::EulerEquation2d)
  @accepts Prefix(lhs), rho, flx_rho_x, flx_rho_y, qx, flx_qx_x, flx_qx_y, qy, flx_qy_y, E, flx_E_x, flx_E_y, v
  @accepts Prefix(rhs), rho, flx_rho_x, flx_rho_y, qx, flx_qx_x, flx_qx_y, qy, flx_qy_y, E, flx_E_x, flx_E_y, v
  @accepts nx, ny

  vmax = absmax(lhs_v, rhs_v)

  lhs_nflx = nx*lhs_flx_rho_x + ny*lhs_flx_rho_y
  rhs_nflx = nx*rhs_flx_rho_x + ny*rhs_flx_rho_y
  nflx_rho = LLF(lhs_nflx, rhs_nflx, lhs_rho, rhs_rho, vmax)

  lhs_nflx = nx*lhs_flx_qx_x + ny*lhs_flx_qx_y
  rhs_nflx = nx*rhs_flx_qx_x + ny*rhs_flx_qx_y
  nflx_qx = LLF(lhs_nflx, rhs_nflx, lhs_qx, rhs_qx, vmax)

  lhs_flx_qy_x = lhs_flx_qx_y
  rhs_flx_qy_x = rhs_flx_qx_y
  lhs_nflx = nx*lhs_flx_qy_x + ny*lhs_flx_qy_y
  rhs_nflx = nx*rhs_flx_qy_x + ny*rhs_flx_qy_y
  nflx_qy = LLF(lhs_nflx, rhs_nflx, lhs_qy, rhs_qy, vmax)

  lhs_nflx = nx*lhs_flx_E_x + ny*lhs_flx_E_y
  rhs_nflx = nx*rhs_flx_E_x + ny*rhs_flx_E_y
  nflx_E = LLF(lhs_nflx, rhs_nflx, lhs_E, rhs_E, vmax)

  @returns nflx_rho, nflx_qx, nflx_qy, nflx_E
end


@with_signature function bdryllf(equation::EulerEquation2d)
  @accepts Prefix(lhs), rho, flx_rho_x, flx_rho_y,
                        qx,  flx_qx_x, flx_qx_y,
                        qy,  #=flx_qy_x,=# flx_qy_y,
                        E,   flx_E_x, flx_E_y, v
  @accepts init_rho, init_qx, init_qy, init_E, init_v
  @accepts nx, ny

  lhs_flx_qy_x = lhs_flx_qx_y

  # enforce dirichlet/neumann conditions following stability analysis
  # see dg-notes/tex/evm.pdf
  # Hesthaven, Numerical Methods for Conservation Laws uses same logic

  # Dirichlet conditions
  rhs_rho = -lhs_rho + 2*init_rho
  rhs_qx  = -lhs_qx  + 2*init_qx
  rhs_qy  = -lhs_qy  + 2*init_qy
  rhs_E   = -lhs_E   + 2*init_E

  # Neumann conditions
  # rhs_rho = lhs_rho
  # rhs_qx  = lhs_qx
  # rhs_qy  = lhs_qy
  # rhs_E   = lhs_E

  error()
  @unpack eos = equation
  rhs_q = sqrt(rhs_qx^2 + rhs_qy^2)
  rhs_eps, = internal_energy((rhs_rho,rhs_qx,rhs_qy,rhs_E), equation)
  rhs_p = eos(Pressure, Density, rhs_rho, InternalEnergy, rhs_eps)
  rhs_flx_rho_x = rhs_qx
  rhs_flx_rho_y = rhs_qy
  rhs_flx_qx_x  = rhs_qx^2 / rhs_rho + rhs_p
  rhs_flx_qx_y  = rhs_qx*rhs_qy / rhs_rho
  rhs_flx_qy_x  = rhs_flx_qx_y
  rhs_flx_qy_y  = rhs_qy^2 / rhs_rho + rhs_p
  rhs_flx_E_x   = rhs_qx / rhs_rho * (rhs_E + rhs_p)
  rhs_flx_E_y   = rhs_qy / rhs_rho * (rhs_E + rhs_p)

  vmax = absmax(lhs_v, init_v)

  lhs_nflx = nx*lhs_flx_rho_x + ny*lhs_flx_rho_y
  rhs_nflx = nx*rhs_flx_rho_x + ny*rhs_flx_rho_y
  nflx_rho = LLF(lhs_nflx, rhs_nflx, lhs_rho, rhs_rho, vmax)

  lhs_nflx = nx*lhs_flx_qx_x + ny*lhs_flx_qx_y
  rhs_nflx = nx*rhs_flx_qx_x + ny*rhs_flx_qx_y
  nflx_qx  = LLF(lhs_nflx, rhs_nflx, lhs_qx, rhs_qx, vmax)

  lhs_nflx = nx*lhs_flx_qy_x + ny*lhs_flx_qy_y
  rhs_nflx = nx*rhs_flx_qy_x + ny*rhs_flx_qy_y
  nflx_qy  = LLF(lhs_nflx, rhs_nflx, lhs_qy, rhs_qy, vmax)

  lhs_nflx = nx*lhs_flx_E_x + ny*lhs_flx_E_y
  rhs_nflx = nx*rhs_flx_E_x + ny*rhs_flx_E_y
  nflx_E = LLF(lhs_nflx, rhs_nflx, lhs_E, rhs_E, vmax)

  @returns nflx_rho, nflx_qx, nflx_qy, nflx_E
end


@with_signature function maxspeed(equation::EulerEquation2d)
  @accepts rho, qx, qy, E
  @unpack eos = equation
  q = sqrt(qx^2+qy^2)
  eps, = internal_energy((rho,qx,qy,E), equation)
  p = eos(Pressure, Density, rho, InternalEnergy, eps)
  v = abs(q / p) + sqrt(abs(eos.gamma * p / rho))
  @returns v
end


@with_signature function entropy(equation::EulerEquation2d{EquationOfState.IdealGas})
  @accepts rho, p
  @unpack eos  = equation
  gamma = eos.gamma
  S = rho / (gamma - 1) * log(p / rho^gamma)
  @returns S
end


@with_signature function entropy_flux(equation::EulerEquation2d)
  @accepts rho, qx, qy, S
  @unpack eos  = equation
  flx_S_x = qx / rho * S
  flx_S_y = qy / rho * S
  @returns flx_S_x, flx_S_y
end


@with_signature function ldg_nflux(eq::EulerEquation2d)
  @accepts Prefix(lhs), rho, qx, qy, E
  @accepts Prefix(rhs), rho, qx, qy, E
  @accepts nx, ny

  avg_rho = 0.5 * (lhs_rho + rhs_rho)
  avg_qx  = 0.5 * (lhs_qx + rhs_qx)
  avg_qy  = 0.5 * (lhs_qy + rhs_qy)
  avg_E   = 0.5 * (lhs_E + rhs_E)

  nflx_rho_x = nx * avg_rho
  nflx_rho_y = ny * avg_rho

  nflx_qx_x  = nx * avg_qx
  nflx_qx_y  = ny * avg_qx

  nflx_qy_x  = nx * avg_qy
  nflx_qy_y  = ny * avg_qy

  nflx_E_x   = nx * avg_E
  nflx_E_y   = ny * avg_E

  @returns nflx_rho_x, nflx_rho_y, nflx_qx_x, nflx_qx_y,
           nflx_qy_x, nflx_qy_y, nflx_E_x, nflx_E_y
end
@with_signature function ldg_bdryllf(eq::EulerEquation2d)
  @accepts Prefix(in), rho, qx, qy, E
  @accepts Prefix(out), rho, qx, qy, E
  @accepts nx, ny

  # set exterior state to 0 according to stability analysis
  # see dg-notes/tex/evm.pdf for reasoning
  nflx_rho_x = nx * in_rho
  nflx_rho_y = ny * in_rho

  nflx_qx_x  = nx * in_qx
  nflx_qx_y  = ny * in_qx

  nflx_qy_x  = nx * in_qy
  nflx_qy_y  = ny * in_qy

  nflx_E_x   = nx * in_E
  nflx_E_y   = ny * in_E

  @returns nflx_rho_x, nflx_rho_y, nflx_qx_x, nflx_qx_y,
           nflx_qy_x, nflx_qy_y, nflx_E_x, nflx_E_y
end


@with_signature function av_flux_2d(eq::EulerEquation2d)
  @accepts flx_rho_x, flx_rho_y, ldg_rho_x, ldg_rho_y,
           flx_qx_x, flx_qx_y, ldg_qx_x, ldg_qx_y,
           #=flx_qy_x,=# flx_qy_y, ldg_qy_x, ldg_qy_y,
           flx_E_x, flx_E_y, ldg_E_x, ldg_E_y,
           smoothed_mu

  flx_qy_x = flx_qx_y

  flx_rho_x += smoothed_mu * ldg_rho_x
  flx_rho_y += smoothed_mu * ldg_rho_y

  flx_qx_x  += smoothed_mu * ldg_qx_x
  flx_qx_y  += smoothed_mu * ldg_qx_y

  flx_qy_x  += smoothed_mu * ldg_qy_x
  flx_qy_y  += smoothed_mu * ldg_qy_y

  flx_E_x   += smoothed_mu * ldg_E_x
  flx_E_y   += smoothed_mu * ldg_E_y

  @returns flx_rho_x, flx_rho_y, flx_qx_x, flx_qx_y,
           flx_qy_x, flx_qy_y, flx_E_x, flx_E_y
end


@with_signature function av_nflux_2d(eq::EulerEquation2d)
  @accepts Prefix(lhs), ldg_rho_x, ldg_rho_y, ldg_qx_x, ldg_qx_y, ldg_qy_x, ldg_qy_y, ldg_E_x, ldg_E_y, smoothed_mu
  @accepts Prefix(rhs), ldg_rho_x, ldg_rho_y, ldg_qx_x, ldg_qx_y, ldg_qy_x, ldg_qy_y, ldg_E_x, ldg_E_y, smoothed_mu
  @accepts nflx_rho, nflx_qx, nflx_qy, nflx_E, nx, ny

  lhs_nflx = nx*lhs_ldg_rho_x + ny*lhs_ldg_rho_y
  rhs_nflx = nx*rhs_ldg_rho_x + ny*rhs_ldg_rho_y
  nflx_rho += 0.5 * (lhs_smoothed_mu*lhs_nflx + rhs_smoothed_mu*rhs_nflx)

  lhs_nflx = nx*lhs_ldg_qx_x + ny*lhs_ldg_qx_y
  rhs_nflx = nx*rhs_ldg_qx_x + ny*rhs_ldg_qx_y
  nflx_qx += 0.5 * (lhs_smoothed_mu*lhs_nflx + rhs_smoothed_mu*rhs_nflx)

  lhs_nflx = nx*lhs_ldg_qy_x + ny*lhs_ldg_qy_y
  rhs_nflx = nx*rhs_ldg_qy_x + ny*rhs_ldg_qy_y
  nflx_qy += 0.5 * (lhs_smoothed_mu*lhs_nflx + rhs_smoothed_mu*rhs_nflx)

  lhs_nflx = nx*lhs_ldg_E_x + ny*lhs_ldg_E_y
  rhs_nflx = nx*rhs_ldg_E_x + ny*rhs_ldg_E_y
  nflx_E += 0.5 * (lhs_smoothed_mu*lhs_nflx + rhs_smoothed_mu*rhs_nflx)

  @returns nflx_rho, nflx_qx, nflx_qy, nflx_E
end



function av_nflux_2d(eq::EulerEquation2d, mesh::Mesh2d)

  @unpack ldg_rho_x, ldg_rho_y, ldg_qx_x, ldg_qx_y, ldg_qy_x, ldg_qy_y,
          ldg_E_x, ldg_E_y, smoothed_mu = get_static_variables(mesh.cache)
  @unpack nflx_rho, nflx_qx, nflx_qy, nflx_E, nx, ny = get_bdry_variables(mesh.cache)

  @inbounds for icell in mesh.bulkidxs

    cell_in = mesh.tree.cells[icell]
    bulk_in = cellindices(mesh, icell)
    rng_lhs, rng_rhs, rng_down, rng_up = dg1d.faceindices(mesh, icell)

    if dg1d.has_neighbor(cell_in, Cart2d.Xmin)
      cell_out = cell_in.neighbors[Cart2d.Xmin]
      bulk_out = cellindices(mesh, cell_out)
      for (i, fi) in enumerate(rng_lhs)
        bo = bulk_out[end,i]
        bi = bulk_in[1,i]

        nxi, nyi = nx[fi], ny[fi]

        lhs_nflx = nxi*ldg_rho_x[bi] + nyi*ldg_rho_y[bi]
        rhs_nflx = nxi*ldg_rho_x[bo] + nyi*ldg_rho_y[bo]
        nflx_rho[fi] += 0.5 * (smoothed_mu[bi]*lhs_nflx + smoothed_mu[bo]*rhs_nflx)

        lhs_nflx = nxi*ldg_qx_x[bi] + nyi*ldg_qx_y[bi]
        rhs_nflx = nxi*ldg_qx_x[bo] + nyi*ldg_qx_y[bo]
        nflx_qx[fi] += 0.5 * (smoothed_mu[bi]*lhs_nflx + smoothed_mu[bo]*rhs_nflx)

        lhs_nflx = nxi*ldg_qy_x[bi] + nyi*ldg_qy_y[bi]
        rhs_nflx = nxi*ldg_qy_x[bo] + nyi*ldg_qy_y[bo]
        nflx_qy[fi] += 0.5 * (smoothed_mu[bi]*lhs_nflx + smoothed_mu[bo]*rhs_nflx)

        lhs_nflx = nxi*ldg_E_x[bi] + nyi*ldg_E_y[bi]
        rhs_nflx = nxi*ldg_E_x[bo] + nyi*ldg_E_y[bo]
        nflx_E[fi] += 0.5 * (smoothed_mu[bi]*lhs_nflx + smoothed_mu[bo]*rhs_nflx)
      end
    end

    if dg1d.has_neighbor(cell_in, Cart2d.Xmax)
      cell_out = cell_in.neighbors[Cart2d.Xmax]
      bulk_out = cellindices(mesh, cell_out)
      for (i, fi) in enumerate(rng_rhs)
        bo = bulk_out[1,i]
        bi = bulk_in[end,i]

        nxi, nyi = nx[fi], ny[fi]

        lhs_nflx = nxi*ldg_rho_x[bi] + nyi*ldg_rho_y[bi]
        rhs_nflx = nxi*ldg_rho_x[bo] + nyi*ldg_rho_y[bo]
        nflx_rho[fi] += 0.5 * (smoothed_mu[bi]*lhs_nflx + smoothed_mu[bo]*rhs_nflx)

        lhs_nflx = nxi*ldg_qx_x[bi] + nyi*ldg_qx_y[bi]
        rhs_nflx = nxi*ldg_qx_x[bo] + nyi*ldg_qx_y[bo]
        nflx_qx[fi] += 0.5 * (smoothed_mu[bi]*lhs_nflx + smoothed_mu[bo]*rhs_nflx)

        lhs_nflx = nxi*ldg_qy_x[bi] + nyi*ldg_qy_y[bi]
        rhs_nflx = nxi*ldg_qy_x[bo] + nyi*ldg_qy_y[bo]
        nflx_qy[fi] += 0.5 * (smoothed_mu[bi]*lhs_nflx + smoothed_mu[bo]*rhs_nflx)

        lhs_nflx = nxi*ldg_E_x[bi] + nyi*ldg_E_y[bi]
        rhs_nflx = nxi*ldg_E_x[bo] + nyi*ldg_E_y[bo]
        nflx_E[fi] += 0.5 * (smoothed_mu[bi]*lhs_nflx + smoothed_mu[bo]*rhs_nflx)
      end
    end

    if dg1d.has_neighbor(cell_in, Cart2d.Ymin)
      cell_out = cell_in.neighbors[Cart2d.Ymin]
      bulk_out = cellindices(mesh, cell_out)
      for (i, fi) in enumerate(rng_down)
        bo = bulk_out[i,end]
        bi = bulk_in[i,1]

        nxi, nyi = nx[fi], ny[fi]

        lhs_nflx = nxi*ldg_rho_x[bi] + nyi*ldg_rho_y[bi]
        rhs_nflx = nxi*ldg_rho_x[bo] + nyi*ldg_rho_y[bo]
        nflx_rho[fi] += 0.5 * (smoothed_mu[bi]*lhs_nflx + smoothed_mu[bo]*rhs_nflx)

        lhs_nflx = nxi*ldg_qx_x[bi] + nyi*ldg_qx_y[bi]
        rhs_nflx = nxi*ldg_qx_x[bo] + nyi*ldg_qx_y[bo]
        nflx_qx[fi] += 0.5 * (smoothed_mu[bi]*lhs_nflx + smoothed_mu[bo]*rhs_nflx)

        lhs_nflx = nxi*ldg_qy_x[bi] + nyi*ldg_qy_y[bi]
        rhs_nflx = nxi*ldg_qy_x[bo] + nyi*ldg_qy_y[bo]
        nflx_qy[fi] += 0.5 * (smoothed_mu[bi]*lhs_nflx + smoothed_mu[bo]*rhs_nflx)

        lhs_nflx = nxi*ldg_E_x[bi] + nyi*ldg_E_y[bi]
        rhs_nflx = nxi*ldg_E_x[bo] + nyi*ldg_E_y[bo]
        nflx_E[fi] += 0.5 * (smoothed_mu[bi]*lhs_nflx + smoothed_mu[bo]*rhs_nflx)
      end
    end

    if dg1d.has_neighbor(cell_in, Cart2d.Ymax)
      cell_out = cell_in.neighbors[Cart2d.Ymax]
      bulk_out = cellindices(mesh, cell_out)
      for (i, fi) in enumerate(rng_up)
        bo = bulk_out[i,1]
        bi = bulk_in[i,end]

        nxi, nyi = nx[fi], ny[fi]

        lhs_nflx = nxi*ldg_rho_x[bi] + nyi*ldg_rho_y[bi]
        rhs_nflx = nxi*ldg_rho_x[bo] + nyi*ldg_rho_y[bo]
        nflx_rho[fi] += 0.5 * (smoothed_mu[bi]*lhs_nflx + smoothed_mu[bo]*rhs_nflx)

        lhs_nflx = nxi*ldg_qx_x[bi] + nyi*ldg_qx_y[bi]
        rhs_nflx = nxi*ldg_qx_x[bo] + nyi*ldg_qx_y[bo]
        nflx_qx[fi] += 0.5 * (smoothed_mu[bi]*lhs_nflx + smoothed_mu[bo]*rhs_nflx)

        lhs_nflx = nxi*ldg_qy_x[bi] + nyi*ldg_qy_y[bi]
        rhs_nflx = nxi*ldg_qy_x[bo] + nyi*ldg_qy_y[bo]
        nflx_qy[fi] += 0.5 * (smoothed_mu[bi]*lhs_nflx + smoothed_mu[bo]*rhs_nflx)

        lhs_nflx = nxi*ldg_E_x[bi] + nyi*ldg_E_y[bi]
        rhs_nflx = nxi*ldg_E_x[bo] + nyi*ldg_E_y[bo]
        nflx_E[fi] += 0.5 * (smoothed_mu[bi]*lhs_nflx + smoothed_mu[bo]*rhs_nflx)
      end
    end
  end

end


# @with_signature function llf(equation::EulerEquation2d)
#   @accepts Prefix(lhs), rho, flx_rho_x, flx_rho_y, qx, flx_qx_x, flx_qx_y, qy, flx_qy_y, E, flx_E_x, flx_E_y, v
#   @accepts Prefix(rhs), rho, flx_rho_x, flx_rho_y, qx, flx_qx_x, flx_qx_y, qy, flx_qy_y, E, flx_E_x, flx_E_y, v
#   @accepts nx, ny
#
#   vmax = absmax(lhs_v, rhs_v)
#
#   lhs_nflx = nx*lhs_flx_rho_x + ny*lhs_flx_rho_y
#   rhs_nflx = nx*rhs_flx_rho_x + ny*rhs_flx_rho_y
#   nflx_rho = LLF(lhs_nflx, rhs_nflx, lhs_rho, rhs_rho, vmax)
#
#   lhs_nflx = nx*lhs_flx_qx_x + ny*lhs_flx_qx_y
#   rhs_nflx = nx*rhs_flx_qx_x + ny*rhs_flx_qx_y
#   nflx_qx = LLF(lhs_nflx, rhs_nflx, lhs_qx, rhs_qx, vmax)
#
#   lhs_flx_qy_x = lhs_flx_qx_y
#   rhs_flx_qy_x = rhs_flx_qx_y
#   lhs_nflx = nx*lhs_flx_qy_x + ny*lhs_flx_qy_y
#   rhs_nflx = nx*rhs_flx_qy_x + ny*rhs_flx_qy_y
#   nflx_qy = LLF(lhs_nflx, rhs_nflx, lhs_qy, rhs_qy, vmax)
#
#   lhs_nflx = nx*lhs_flx_E_x + ny*lhs_flx_E_y
#   rhs_nflx = nx*rhs_flx_E_x + ny*rhs_flx_E_y
#   nflx_E = LLF(lhs_nflx, rhs_nflx, lhs_E, rhs_E, vmax)
#
#   @returns nflx_rho, nflx_qx, nflx_qy, nflx_E
# end
function llf(equation::EulerEquation2d, mesh::Mesh2d)
  @unpack rho, qx, qy, E = get_dynamic_variables(mesh.cache)
  @unpack flx_rho_x, flx_rho_y, flx_qx_x, flx_qx_y, flx_qy_y,
          flx_E_x, flx_E_y, v = get_static_variables(mesh.cache)
  @unpack nx, ny, nflx_rho, nflx_qx, nflx_qy, nflx_E = get_bdry_variables(mesh.cache)

  flx_qy_x = flx_qx_y

  @inbounds for icell in mesh.bulkidxs

    cell_in = mesh.tree.cells[icell]
    bulk_in = cellindices(mesh, icell)
    rng_lhs, rng_rhs, rng_down, rng_up = dg1d.faceindices(mesh, icell)

    if dg1d.has_neighbor(cell_in, Cart2d.Xmin)
      cell_out = cell_in.neighbors[Cart2d.Xmin]
      bulk_out = cellindices(mesh, cell_out)
      for (i, fi) in enumerate(rng_lhs)
        bo = bulk_out[end,i]
        bi = bulk_in[1,i]

        vmax = absmax(v[bi], v[bo])

        nxi, nyi = nx[fi], ny[fi]

        lhs_nflx = nxi*flx_rho_x[bi] + nyi*flx_rho_y[bi]
        rhs_nflx = nxi*flx_rho_x[bo] + nyi*flx_rho_y[bo]
        nflx_rho[fi] = LLF(lhs_nflx, rhs_nflx, rho[bi], rho[bo], vmax)

        lhs_nflx = nxi*flx_qx_x[bi] + nyi*flx_qx_y[bi]
        rhs_nflx = nxi*flx_qx_x[bo] + nyi*flx_qx_y[bo]
        nflx_qx[fi] = LLF(lhs_nflx, rhs_nflx, qx[bi], qx[bo], vmax)

        lhs_nflx = nxi*flx_qy_x[bi] + nyi*flx_qy_y[bi]
        rhs_nflx = nxi*flx_qy_x[bo] + nyi*flx_qy_y[bo]
        nflx_qy[fi] = LLF(lhs_nflx, rhs_nflx, qy[bi], qy[bo], vmax)

        lhs_nflx = nxi*flx_E_x[bi] + nyi*flx_E_y[bi]
        rhs_nflx = nxi*flx_E_x[bo] + nyi*flx_E_y[bo]
        nflx_E[fi] = LLF(lhs_nflx, rhs_nflx, E[bi], E[bo], vmax)

      end
    end

    if dg1d.has_neighbor(cell_in, Cart2d.Xmax)
      cell_out = cell_in.neighbors[Cart2d.Xmax]
      bulk_out = cellindices(mesh, cell_out)
      for (i, fi) in enumerate(rng_rhs)
        bo = bulk_out[1,i]
        bi = bulk_in[end,i]

        vmax = absmax(v[bi], v[bo])

        nxi, nyi = nx[fi], ny[fi]

        lhs_nflx = nxi*flx_rho_x[bi] + nyi*flx_rho_y[bi]
        rhs_nflx = nxi*flx_rho_x[bo] + nyi*flx_rho_y[bo]
        nflx_rho[fi] = LLF(lhs_nflx, rhs_nflx, rho[bi], rho[bo], vmax)

        lhs_nflx = nxi*flx_qx_x[bi] + nyi*flx_qx_y[bi]
        rhs_nflx = nxi*flx_qx_x[bo] + nyi*flx_qx_y[bo]
        nflx_qx[fi] = LLF(lhs_nflx, rhs_nflx, qx[bi], qx[bo], vmax)

        lhs_nflx = nxi*flx_qy_x[bi] + nyi*flx_qy_y[bi]
        rhs_nflx = nxi*flx_qy_x[bo] + nyi*flx_qy_y[bo]
        nflx_qy[fi] = LLF(lhs_nflx, rhs_nflx, qy[bi], qy[bo], vmax)

        lhs_nflx = nxi*flx_E_x[bi] + nyi*flx_E_y[bi]
        rhs_nflx = nxi*flx_E_x[bo] + nyi*flx_E_y[bo]
        nflx_E[fi] = LLF(lhs_nflx, rhs_nflx, E[bi], E[bo], vmax)
      end
    end

    if dg1d.has_neighbor(cell_in, Cart2d.Ymin)
      cell_out = cell_in.neighbors[Cart2d.Ymin]
      bulk_out = cellindices(mesh, cell_out)
      for (i, fi) in enumerate(rng_down)
        bo = bulk_out[i,end]
        bi = bulk_in[i,1]

        vmax = absmax(v[bi], v[bo])

        nxi, nyi = nx[fi], ny[fi]

        lhs_nflx = nxi*flx_rho_x[bi] + nyi*flx_rho_y[bi]
        rhs_nflx = nxi*flx_rho_x[bo] + nyi*flx_rho_y[bo]
        nflx_rho[fi] = LLF(lhs_nflx, rhs_nflx, rho[bi], rho[bo], vmax)

        lhs_nflx = nxi*flx_qx_x[bi] + nyi*flx_qx_y[bi]
        rhs_nflx = nxi*flx_qx_x[bo] + nyi*flx_qx_y[bo]
        nflx_qx[fi] = LLF(lhs_nflx, rhs_nflx, qx[bi], qx[bo], vmax)

        lhs_nflx = nxi*flx_qy_x[bi] + nyi*flx_qy_y[bi]
        rhs_nflx = nxi*flx_qy_x[bo] + nyi*flx_qy_y[bo]
        nflx_qy[fi] = LLF(lhs_nflx, rhs_nflx, qy[bi], qy[bo], vmax)

        lhs_nflx = nxi*flx_E_x[bi] + nyi*flx_E_y[bi]
        rhs_nflx = nxi*flx_E_x[bo] + nyi*flx_E_y[bo]
        nflx_E[fi] = LLF(lhs_nflx, rhs_nflx, E[bi], E[bo], vmax)
      end
    end

    if dg1d.has_neighbor(cell_in, Cart2d.Ymax)
      cell_out = cell_in.neighbors[Cart2d.Ymax]
      bulk_out = cellindices(mesh, cell_out)
      for (i, fi) in enumerate(rng_up)
        bo = bulk_out[i,1]
        bi = bulk_in[i,end]

        vmax = absmax(v[bi], v[bo])

        nxi, nyi = nx[fi], ny[fi]

        lhs_nflx = nxi*flx_rho_x[bi] + nyi*flx_rho_y[bi]
        rhs_nflx = nxi*flx_rho_x[bo] + nyi*flx_rho_y[bo]
        nflx_rho[fi] = LLF(lhs_nflx, rhs_nflx, rho[bi], rho[bo], vmax)

        lhs_nflx = nxi*flx_qx_x[bi] + nyi*flx_qx_y[bi]
        rhs_nflx = nxi*flx_qx_x[bo] + nyi*flx_qx_y[bo]
        nflx_qx[fi] = LLF(lhs_nflx, rhs_nflx, qx[bi], qx[bo], vmax)

        lhs_nflx = nxi*flx_qy_x[bi] + nyi*flx_qy_y[bi]
        rhs_nflx = nxi*flx_qy_x[bo] + nyi*flx_qy_y[bo]
        nflx_qy[fi] = LLF(lhs_nflx, rhs_nflx, qy[bi], qy[bo], vmax)

        lhs_nflx = nxi*flx_E_x[bi] + nyi*flx_E_y[bi]
        rhs_nflx = nxi*flx_E_x[bo] + nyi*flx_E_y[bo]
        nflx_E[fi] = LLF(lhs_nflx, rhs_nflx, E[bi], E[bo], vmax)
      end
    end
  end
end
