function get_outputdir()
  thisdir = dirname(@__FILE__)
  outputdir = "$thisdir/../outputs/"
  outputdir = normpath(outputdir)
  if !isdir(outputdir)
    mkdir(outputdir)
  end
  return outputdir
end


@deprecate get_output_dir() get_outputdir()


# stolen from https://discourse.julialang.org/t/convert-type-to-dictionary/6982/7
typedict(x::T) where T = Dict(fn=>getfield(x,fn) for fn ∈ fieldnames(T))


dropmissing(collection) = collect(skipmissing(collection))
function dropmissing(ntpl::NamedTuple)
  cleaned_ntpl = []
  for (key, val) in zip(keys(ntpl), values(ntpl))
    if !ismissing(val)
      push!(cleaned_ntpl, key => val)
    end
  end
  return NamedTuple(cleaned_ntpl)
end



function query_key(prms, key, default)
  value = get(prms, key, missing)
  if ismissing(value)
    return default
  elseif value isa String
    return Sandbox.parse_expression_string(value)
  else
    return value
  end
end


"""
    query(parameters, key; default=missing, options=[])
    query(T, parameters, key; default=missing, options=[]) where T

Query the value referenced by `key` from `parameters` and return `default` if nothing is found.
Use `options` to restrict the allowed values, where the first match will be returned.
Note: `options` must be a vector of homogenuous types and, if specified, must
be compatible with the type parameter `T`.
"""
function query(::Type{T}, parameters, key;
    default=missing, options=[], errormsg="") where T<:Union{Number,AbstractString,AbstractVector}

  value = query(parameters, key; default=default, options=options, errormsg=errormsg)
  if T <: Integer && value isa AbstractFloat
    value = round(value)
  end
  try
    return convert(T, value)
  catch e
    error("Conversion of value '$value' to type '$T' for key '$key' failed.")
  end
end


function query(parameters, key;
    default=missing, options=[], errormsg::String="")

  value = query_key(parameters, key, default)
  unique_options = unique(options)
  length(unique_options) != length(options) && error("'options' must be unique")
  length(options) == 0 && return value
  # match against options
  T = eltype(options)
  if T <: Union{Number, AbstractVector, AbstractString}
    if !(value in options)
      if ismissing(default)
        error("Found value '$value' not matching options '$(join(options,", ")...)'")
      end
      return default
    end
    return value
  else
    error("Cannot match against options of type '$T'")
  end
end


"""
    query(d::AbstractDict, rgx::Regex)

Query a subset from `d` for which keys match the regex pattern `rgx`.
"""
function query(d::T, rgx::Regex) where T<:AbstractDict
  d_keys     = collect(keys(d))
  d_values   = collect(values(d))
  rgx_matches = match.(rgx, d_keys)
  matched_indices = map(k -> !isnothing(k), rgx_matches)
  matched_keys = d_keys[matched_indices]
  matched_values = d_values[matched_indices]
  rgx_d = T( (k,v) for (k,v) in zip(matched_keys, matched_values) )
end


"""
    trim_keys(dict::Dict{T,S}, subkey::P) where {T,S,P<:Union{String,Symbol}}

Trim pattern `subkey` from the start of all keys of dictionary `dict`.
"""
function trim_keys(dict::AbstractDict, subkey)
  length(dict) == 0 && return dict
  ktype, vtype = keytype(dict), valtype(dict)
  newdict = Dict{ktype,vtype}()
  str_subkey = String(subkey)
  L = length(str_subkey)
  for (key, value) in pairs(dict)
    str_key = String(key)
    if length(str_key) == L 
      error("Failed to remove subkey '$subkey' from key '$key', because that would result in " *
            "an empty key")
    end
    if length(str_key) < L || str_key[1:L] != str_subkey
      error("Failed to remove subkey '$subkey' from key '$key'")
    end
    newdict[ktype(str_key[L+1:end])] = value
  end
  return newdict
end


"""
    match_keys(fn::Function, dict::Dict{T,S}) where {T,S}

Match keys using a function `fn(k)::Bool`, return matched and unmatched parts in separate 
dictionaries.
"""
function match_keys(fn::Function, dict::T) where T<:AbstractDict
  keys_dict, values_dict = collect(keys(dict)), collect(values(dict))
  matched_indices   = map(k -> !isnothing(fn(k)), keys_dict)
  matched_keys      = keys_dict[matched_indices]
  matched_values    = values_dict[matched_indices]
  unmatched_indices = (!).(matched_indices) # broadcast !
  unmatched_keys    = keys_dict[unmatched_indices]
  unmatched_values  = values_dict[unmatched_indices]
  matched_dict = T( (k,v) for (k,v) in zip(matched_keys, matched_values) )
  unmatched_dict = T( (k,v) for (k,v) in zip(unmatched_keys, unmatched_values) )
  return matched_dict, unmatched_dict
end


function prepend_keys(dict::T, header) where T<:AbstractDict
  ktype = keytype(dict)
  return T( (ktype("$(string(header)).$(string(k))"), v) for (k,v) in pairs(dict) )
end


"""
  concretetypes(type)

Gather all concrete subtypes of a `type` by applying `subtypes` recursively.
"""
function concretetypes(type)
  ctypes = []
  for sti in subtypes(type)
    if isabstracttype(sti)
      append!(ctypes, concretetypes(sti))
    else
      push!(ctypes, sti)
    end
  end
  return ctypes
end


"""
    splitallext(path, ext="")

Remove all extensions from variable `path`.
To remove only the last extension use `Base.splitext`.
"""
function splitallext(path, ext="")
  newpath, newext = splitext(path)
  if length(newext) == 0
    return (newpath, ext)
  else
    newext = join([newext, ext])
    return splitallext(newpath, newext)
  end
end


"""
    mk_outputdir(path)

Generate an output directory in root directory of `path`. Directory is only created if
`path` is a .toml file.
If a folder with the same name is already present, then it is backed up by moving it
and appending .previous to its name.
"""
function mk_outputdir(path)
  path = abspath(path)

  # analyze extensions
  basename, allext = splitallext(path)
  allext = split(allext, '.', keepempty=false)
  if length(allext) == 0
    # assume toml file
    allext = [".toml"]
  end
  firstexts, lastext = if length(allext) == 1
    [], allext[1]
  else
    allext[1:end-1], allext[end]
  end

  lastext == "toml" || error("Only accept parameter files in .toml format, received '$path'")

  isfile(path) || error("Parameter file '$path' not found")

  outputdirname = basename
  if length(firstexts) > 0
    outputdirname = "$basename.$(join(firstexts, '.'))"
  end
  # backup if there is already an output folder
  if isdir(outputdirname)
    @info("Backing up '$outputdirname'")
    mv(outputdirname, "$(outputdirname).previous", force=true)
  end
  mkdir(outputdirname, mode=0o775)
  return outputdirname
end


"""
    flatten(d::AbstractDict{T,<:Any}, prefix_delim = ".") where T

Flatten a dictionary `d` and concatenate all keys with `prefix_delim`.
Returned dictionary is of type `Dict{Symbol,<:Any}`.
"""
function flatten(d::AbstractDict{T,<:Any}, prefix_delim = ".") where T
  new_d = Dict{T,Any}()
  for (key, value) in pairs(d)
    if isa(value, Dict)
      flattened_value = flatten(value, prefix_delim)
      for (ikey, ivalue) in pairs(flattened_value)
        new_d["$key.$ikey"] = ivalue
      end
    else
      new_d[key] = value
    end
  end
  return new_d
end


"""
    dirname_path(path)

Return directory basename of `path`.
"""
function dirname_path(path)
  if !ispath(path)
    error("Could not locate a file or directory at path '$path'")
  end
  tmp_path = abspath(normpath(path))
  basename_dir = if isfile(tmp_path)
    basename(dirname(tmp_path))
  else
    if tmp_path[end] == '/'
      tmp_path = tmp_path[1:end-1]
    end
    basename(tmp_path)
  end
  if length(basename_dir) == 0
    error("Failed to retrieve basename from path '$path'")
  end
  return basename_dir
end


function get_parameter_file(dir)
  dir = abspath(dir)
  @assert isdir(dir) "'$dir' is not a directory name"
  dirname = dirname_path(dir)
  filename = joinpath(dir, "$(dirname).toml")
  isfile(filename) || error("Could not find parameter file '$filename'")
  return basename(filename)
end


function get_parameter_file_index(filename)
  basefilename, allexts = splitallext(filename)
  rgx = r".*_(\d+)$"
  result = match(rgx, basefilename)
  if isnothing(result)
    error("Could not extract an index from file '$filename'")
  end
  return first(result.captures)
end


"""
    gather_output_directories(projectdir)

Scan `projectdir` for directories named `<projectdir>_###` and return matches.
"""
function gather_output_directories(projectdir)
  projectdir = normpath(projectdir)
  projectdirname = dirname_path(projectdir)
  pattern = "$(projectdirname)_*"
  dirs = glob(pattern, projectdir)
  filter!(isdir, dirs)
  filter!(d->match(r".*\.previous",d) === nothing, dirs) # filter backups
  return dirs
end


"""
    gather_parameter_files(projectdir)

Scan `projectdir` for `.toml` files named `<projectdir>_###.toml` and return matches.
"""
function gather_parameter_files(projectdir)
  projectdir = abspath(projectdir)
  projectdirname = dirname_path(projectdir)
  pattern = "$(projectdirname)_*.toml"
  files = glob(pattern, projectdir)
  filter!(isfile, files)
  return files
end


"""
    select_parameter_files(projectdir)

Interactively select parameter files from `projectdir`.
"""
function select_parameter_files(dir; display_basename=true)
  dir = abspath(normpath(dir))
  !isdir(dir) && error("'$dir' is not a directory")
  files = gather_parameter_files(dir)
  display_files = display_basename ? basename.(files) : files
  TerminalMenus.Config(; charset=:unicode, scroll_wrap=true)
  menu = TerminalMenus.MultiSelectMenu(display_files)
  selected = TerminalMenus.request("Select parameter files:", menu)
  return files[collect(selected)]
end


"""
    find_duplicated_parameter_files(projectdir)

Scan `projectdir` for `.toml` files with duplicated parameters and return a list of any matches.
"""
function find_duplicated_parameter_files(projectdir)
  projectdir = normpath(projectdir)
  projectdirname = dirname_path(projectdir)
  parameterfiles = gather_parameter_files(projectdir)
  parameters = [ flatten(TOML.parsefile(file)) for file in parameterfiles ]
  list_duplicated_files = []
  for (idx, (file, pars)) in enumerate(zip(parameterfiles, parameters))
    # skip if duplication already detected
    any([ file in duplicates for duplicates in list_duplicated_files ]) && continue
    indices = findall(p -> p == pars, parameters)
    if length(indices) > 1
      push!(list_duplicated_files, parameterfiles[indices])
    end
  end
  return list_duplicated_files
end


"""
    run_projects(main_fn, projectdir, threaded::Bool=true, force::Bool=false)

Run `main_fn` for all parameter files found in `projectdir` with name `projectdir_###.toml`.
If `threaded = true`, run in parallel (does not work with Makie).
TODO: Implement this: If `force=false`, only run parameter files with recent changes.
"""
function run_projects(fn, projectdir; threaded::Bool=false, force::Bool=true)
  nthreads = Threads.nthreads()

  tomls = gather_parameter_files(projectdir)
  if !force
    error("TODO: Figure out recently changed parameter files if force=false.")
  end

  # TODO It is not possible to extract the worker and write an if to switch between
  # threaded and non-threaded loops, because the body contains a break.
  # Is there a way around to wrap the @threads macro in an if to collapse this body?
  # Probably not, because macros are evaluated at compile time ...
  if threaded && nthreads > 1
    println("Running with $(Threads.nthreads()) threads ...")
    Threads.@threads for toml in tomls
      @info "Processing '$toml' ..."
      try
        fn(toml)
      catch e;
        if e isa InterruptException
          break
        elseif e isa AssertionError
          @info e
        else
          rethrow(e)
        end
      end
    end
  else
    for toml in tomls
      @info "Processing '$toml' ..."
      try
        fn(toml)
      catch e
        if e isa InterruptException
          break
        elseif e isa AssertionError
          @info e
        else
          rethrow(e)
        end
      end
    end
  end
end


"""
    analyze_projects(analyze_fn, projectdir, threaded::Bool=true, force::Bool=false)

Run `analyze_fn` for all output directories found in `projectdir`.
TODO: Implement this: If `force=false`, only process directories that were recently changed.
"""
function analyze_projects(analyze_fn, projectdir; force::Bool=false)
  nthreads = Threads.nthreads()

  dirs = gather_output_directories(projectdir)

  for dir in dirs
    @info "Processing '$dir' ..."
    try
      analyze_fn(dir)
    catch e
      if e isa InterruptException
        break
      elseif e isa AssertionError
        @info e
      else
        rethrow(e)
      end
    end
  end
end


#######################################################################
#                             DEPRECATED                              #
#######################################################################


function meshdict(; basedict=Dict(), kwargs...)
  basedict = mergewith(+, Dict{String,Any}(), basedict) # widden type

  # determine which kwargs we should mesh =^= the ones for which a vector of elements is provided
  # kwargs with a single value are just added to the base dict
  # TODO: Remove the single value arguments and assume meshing wrt all kwargs. E.g. any specialized
  # values should be added to the basedict before meshing.
  mixargs = []
  for (key, val) in kwargs
    if !(val isa AbstractArray)
      basedict[string(key)] = val
    else
      push!(mixargs, (key,val))
    end
  end

  # specialize basedict for every collection of values provided
  mixkeys, mixvals = [ m[1] for m in mixargs ], [ m[2] for m in mixargs ]
  allmixes = mix(mixvals...)
  ds = Dict{String,Any}[]
  for mix in allmixes
    tmpdict = copy(basedict)
    for (i, k) in enumerate(mixkeys)
      tmpdict[string(k)] = mix[i]
    end
    push!(ds, tmpdict)
  end
  return ds
end


# reshape allocates, see https://github.com/JuliaLang/julia/issues/36313
# however, Base allows to construct a custom reshape that does not allocate
@inline vreshape(x, sz::Tuple{N,Int64}) where N = reshape(view(x, :), sz)
@inline rview(x, rng, sz) = reshape(view(x, rng), sz)


TODO() = error("Not implemented yet!")
TODO(msg::String) = error(msg)
TODO(fn::Function) = error("'$fn': Not implemented yet!")
TODO(fn::Function, msg) = error("""
'$fn'' Not implemented yet!
$msg
""")

macro TODO()
  :(error("Not implemented yet!"))
end


"""
    @stringify_names x
    @stringify_names [ x, y, z ]

Returns the initial argument as well as stringified variable names in the same storage layout.

# Examples

```julia
julia> x = 1;

julia> @stringify_names x
(x, :x)

julia> y, z = 2, 3;

julia> @stringify_names [ x, y, z ]
([1, 2, 3], ["x", "y", "z"])
```
"""
macro stringify_names(expr)
  str_expr = if expr.head === :vect
    str_elements = map(expr.args) do a
      !isnothing(a) ? string(a) : nothing
    end
    Expr(:vect, str_elements...)
  elseif expr.head === Symbol
    expr.args[1]
  else
    error("Cannot stringify expression type '$(expr.head)' yet!")
  end

  return Expr(:tuple, esc(expr), str_expr)
end


"""
    indent(s::String; level=0, fillwith='\t')

Prepand all lines of string `s` exactly `level` times with `fillwith`.

# Example

```
julia> s1 = \"""
            Level 1
              a = 1
            \"""

julia> print(indent(s1, level=2, fillwith="***"))
******Level 1
******  a = 1
```
"""
function indent(s::String; level=0, fillwith='\t')
  newlines = findall(c->c=='\n', s)
  n = length(newlines)
  indented_s = repeat(fillwith, level)
  for idx = 1:n
    nline      = newlines[idx]
    prev_nline = idx > 1 ? newlines[idx-1]+1 : 1
    next_nline = idx < n ? newlines[idx] : n
    indented_s *= s[prev_nline:nline]
    if idx < n
      indented_s *= repeat(fillwith, level) * s[nline+1:next_nline]
    end
  end
  return indented_s
end
